"""
We write examples in the test suite so that we don't have to use doctests.
Doctests are handy for simple and limited examples, but have their own issues
such as: (1) cannot correctly run or detect doctests separated by text or (2)
annoying clutter (>>> or ...) that hinders readability.

References
----------
https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-literalinclude
https://stackoverflow.com/questions/77172274
"""