"""By default, pytest runs only notebooks from \examples.
- \examples are lightweight to run (i.e. not computationally expensive)
and their main purpose is to show the trading-gym api.
- \applications are real-world application, which can be computationally
expensive to run.
"""